/* ------------------------------------------------------------------------
 * $Id: VNCDesktop.cc,v 1.2 2001/08/22 12:23:20 elm Exp $
 *
 * This file is part of 3Dwm: The Three-Dimensional User Environment.
 *
 * 3Dwm: The Three-Dimensional User Environment:
 *	<http://www.3dwm.org>
 *
 * Chalmers Medialab
 * 	<http://www.medialab.chalmers.se>
 * 
 * ------------------------------------------------------------------------
 * File created 2000-10-25 by Steve Houston.
 *
 * Copyright (c) 2000, 2001 Niklas Elmqvist <elm@3dwm.org>.
 * Copyright (c) 2000 Steve Houston <shouston@programmer.net>.
 * ------------------------------------------------------------------------
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * ------------------------------------------------------------------------
 */

// -- System includes

// -- 3Dwm includes
#include "Nobel/Mapper.hh"
#include "Celsius/Mutex.hh"
#include "Celsius/Guard.hh"
#include "Garbo/VNCInterface.hh"
#include "Polhem/VNCDesktop.hh"
#include "Celsius/Thread.hh"

using namespace Nobel;

// -- Code Segment
VNCTextureGenerator::VNCTextureGenerator(VNCInterface *vnc)
    : _vnc(vnc)
{    
    // Get the first update
    _vnc->updateFramebuffer(0, 0, getWidth(), getHeight());
}

VNCTextureGenerator::~VNCTextureGenerator()
{
    // empty
}

CORBA::Long VNCTextureGenerator::getWidth() 
{
    return (CORBA::Long) _vnc->getWidth(); 
}

CORBA::Long VNCTextureGenerator::getHeight() 
{
    return (CORBA::Long) _vnc->getHeight(); 
}

CORBA::Short VNCTextureGenerator::getPixelSize() 
{
    return (CORBA::Short) _vnc->getPixelSize(); 
}

Nobel::PixelType VNCTextureGenerator::getPixelType() 
{ 
    return _vnc->getPixelType(); 
}

BinaryData* VNCTextureGenerator::getTextureData()
{
    Guard<Mutex> guard(_mutex);
    int bufSize;
    
    // Retrieve the framebuffer data from the VNC interface
    CORBA::Octet *buf = (CORBA::Octet *) _vnc->getFramebuffer();

    // Retain control of memory management. Callee must free
    // BinaryData when finished.
    bufSize = getHeight() * getWidth() * (getPixelSize() / 8);
    BinaryData *texData = new BinaryData(bufSize, bufSize, buf, 0);
    return texData;
}

CORBA::Boolean VNCTextureGenerator::getUpdate(BinaryData_out texData,
					      Rectangle &rect)
{
    Guard<Mutex> guard(_mutex);
    WindowingSystemInterface::Rectangle r;
    bool update = false;
    
    // TODO: we should probably be calling into
    // WindowingSystemInterface instead of VNCInterface in case
    // something other than VNC being used to generate framebuffer
    // updates.
    
    while (_vnc->getFramebufferUpdate(r) == true)
	update = true;
    
    if (update == false)
	return false;

    // TODO: for now we are forcing complete screen refresh
    r.x = 0;
    r.y = 0;
    r.width = getWidth();
    r.height = getHeight();

    // Transfer the rectangle info from the rect struct used in WSInterface
    // into the rect struct passed around by CORBA.
    // TODO: Unify these to eliminate this copy.
    rect.x = r.x;
    rect.y = r.y;
    rect.w = r.width;
    rect.h = r.height;

    int updateSize = r.width * r.height * (getPixelSize() / 8);

    CORBA::Octet *buf = (CORBA::Octet *) _vnc->getFramebuffer();

    // Retain control of memory management. Callee must free BinaryData 
    // when finished
    texData = new BinaryData(updateSize, updateSize, buf, 0);
    return true;
}

VNCDesktop::VNCDesktop(const char *host, CORBA::Short port, const char *pwd)
    : _vnc(new VNCInterface(host, port, pwd)),
      _generator(new VNCTextureGenerator(_vnc))
{
    // TODO: throw exception if unable to connect to server
    
    // Create a thread to handle the VNC connection
    _vnc_thread = new Thread(_vnc);
    _vnc_thread->start();
}

VNCDesktop::~VNCDesktop()
{
    // Kill the VNC thread
    _vnc_thread->kill();
    
    // Delete objects
    delete _vnc_thread;
    delete _vnc;
}

DesktopController_ptr VNCDesktop::createController(Node_ptr body)
{
    Guard<Mutex> guard(_mutex);
    VNCController *controller = activate(new VNCController(_vnc));
    controller->setBody(body);
    return controller->_this();
    //return DesktopController::_nil();
}

TextureGenerator_ptr VNCDesktop::getTextureGenerator()
{
    Guard<Mutex> guard(_mutex);
    return _generator->_this();
}

CORBA::Long VNCDesktop::getWidth()
{
    return _vnc->getWidth();    
}

CORBA::Long VNCDesktop::getHeight()
{
    return _vnc->getWidth();    
}

VNCController::VNCController(VNCInterface *vnc)
    : _vnc(vnc) { }

VNCController::~VNCController() { }

void VNCController::setMapper(Mapper_ptr m)
{
    Guard<Mutex> guard(_mutex);
    _mapper = Nobel::Mapper::_duplicate(m);
}

Mapper_ptr VNCController::getMapper()
{
    Guard<Mutex> guard(_mutex);
    return Nobel::Mapper::_duplicate(_mapper);
}

bool VNCController::handleNodeEvents(const Nobel::Event &e)
{
    // If we don't have a mapper, we can't do anything
    if (CORBA::is_nil(_mapper))
	return false;

    // Now, use the mapper to convert to texture and then screen
    // coordinates.
    TexCoord t = _mapper->map(e.pos);
    int x = int(t.u * _vnc->getWidth());
    int y = int((1.0 - t.v) * _vnc->getHeight());
    
    // See if we can handle this event
    switch (e.type) {

    case KeyPress: 
	if (e.dev == 0) {
	    _vnc->keyEvent(e.key, true);
	    _vnc->keyEvent(e.key, false);
	}
	else {
	    _vnc->pointerEvent(x, y, (1 << e.key));
	    _vnc->pointerEvent(x, y, 0);
	}
	return true;

	/*
    case KeyRelease:
	if (e.dev == 0) _vnc->keyEvent(e.key, false);
	else 
	    _vnc->pointerEvent(x, y, 0);
	return true;
	*/
    case Movement:
	_vnc->pointerEvent(x, y, 0);
	return true;
	
    default: break;
    }
    
    // No, can't handle the event
    return false;
}
    
